module L = Llvm
module StringMap = Map.Make(String)

let txt_of_type = function
      Sast.Num      -> "Num"
    | Sast.Bool     -> "Bool"
    | Sast.String   -> "String"
    | Sast.Void     -> "Void"
    | Sast.List(_)  -> "List"
    | _ -> "Other types" 


let map_lst = ref [|Hashtbl.create 100|]
(* struct to llvm struct definition *)
let struct_types = Hashtbl.create 100
(* (struct ^ struct fields) fields to struct index *)
let struct_field_indexes = Hashtbl.create 100
(* global table mapping struct name -> struct pointer *)
let obj_ptr_map = Hashtbl.create 100
(* table for list len *)
let list_len_hash = Hashtbl.create 42

let rec is_new_var s i =
  if i = Array.length !map_lst then true
  else if Hashtbl.mem (Array.get !map_lst i) s then false
  else is_new_var s (i+1)

let rec lookup s i =
    if i = (Array.length !map_lst)
    then
      raise(Failure("Variable not found"))
    else if Hashtbl.mem (Array.get !map_lst i) s 
    then
      Hashtbl.find (Array.get !map_lst i) s
    else lookup s (i+1)

let translate functions obj_hash _ =
  let context     = L.global_context () in
  let the_module  = L.create_module context "Pseudo"
  and f_t         = L.double_type context
  and i32_t       = L.i32_type  context
  and i8_t        = L.i8_type   context
  and i1_t        = L.i1_type   context
  and str_t       = L.pointer_type (L.i8_type context)
  and void_t      = L.void_type context

  in
  let rec ltype_of_typ = function
      Sast.Num   -> f_t 
    | Sast.Bool  ->  i1_t 
    | Sast.String -> str_t
    | Sast.List(t) -> ltype_of_typ t
    | Sast.Void -> void_t
    | Sast.Object(id) -> Hashtbl.find struct_types (string_of_int id)
    | _ -> raise(Failure("Invalid Data Type"))

  and return_typ = function
      Sast.Num -> f_t
    | Sast.Bool -> i1_t
    | Sast.String -> str_t
    | Sast.Void -> void_t
    | Sast.List(t) -> L.pointer_type (ltype_of_typ t)
    | Sast.Object(id) -> L.pointer_type (Hashtbl.find struct_types (string_of_int id))
    | _ -> raise(Failure("Invalid Data Type"))
  in

  (* each function has its own StringMap, each scope needs one too *)
  let printf_t = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func = L.declare_function "printf" printf_t the_module in
  let prints_t = L.var_arg_function_type str_t [| L.pointer_type i8_t |] in
  let prints_func = L.declare_function "puts" prints_t the_module in
  let printss_t = L.var_arg_function_type str_t [| L.pointer_type i8_t |] in
  let printss_func = L.declare_function "printf" printss_t the_module in

  let codegen_struct c =
    let struct_t = L.named_struct_type context (string_of_int c) in
    let _ = Hashtbl.add struct_types (string_of_int c) struct_t in
    let field_list = Hashtbl.find obj_hash c in
    let type_list = List.map(fun(_, t) -> let x = (match t with 
                                            Sast.List(t) -> L.pointer_type (ltype_of_typ t)
                                          | Sast.Object(_) -> L.pointer_type (ltype_of_typ t)
                                          | _ -> ltype_of_typ t) in x) field_list in 
    let name_list = List.map (fun (s, _) -> s) field_list in

    let type_array = (Array.of_list type_list) in
    List.iteri (fun i f ->
      let n = (string_of_int c) ^ "." ^ f in
      Hashtbl.add struct_field_indexes n i;
    ) name_list;
    L.struct_set_body struct_t type_array false
  in
  let _ = Hashtbl.iter (fun k _ -> codegen_struct k) obj_hash in
  let function_decls =
    let function_decl m fdecl =
      let name = fdecl.Sast.afname 
      and formal_types =
        Array.of_list (List.map (fun (t, _) -> 
          (match t with
            | Sast.List(t') -> L.pointer_type (ltype_of_typ t')
            | Sast.Object(_) -> L.pointer_type (ltype_of_typ t)
            | _ -> ltype_of_typ t)) 
      fdecl.Sast.aformals) in
      let ftype = L.function_type (return_typ fdecl.Sast.return) formal_types in
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in
    List.fold_left function_decl StringMap.empty functions in

  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.Sast.afname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and float_format_str = L.build_global_stringptr "%.02f\n" "fmt" builder

    and string_list_iter = L.build_global_stringptr "%s, " "fmt" builder
    and string_list_ender = L.build_global_stringptr "%s" "fmt" builder

    and int_list_iter = L.build_global_stringptr "%d, " "fmt" builder
    and int_list_ender = L.build_global_stringptr "%d" "fmt" builder

    and float_list_iter = L.build_global_stringptr "%.02f, " "fmt" builder
    and float_list_ender = L.build_global_stringptr "%.02f" "fmt" builder in

    let add_formal (t,n) p = L.set_value_name n p;
      let ret = (match t with
        | Sast.Object(_) -> p
        | Sast.List(_) -> p
        | _ -> let local = L.build_alloca (ltype_of_typ t) n builder in
                let _ = L.build_store p local builder in 
                  local)
      in Hashtbl.add (Array.get !map_lst 0) n ret;
    in
    let _ = List.iter2 add_formal fdecl.Sast.aformals (Array.to_list (L.params the_function)) in
  (* Return the value for a variable or formal argument *)
   let add_var s t builder =
    let found_var_arr = Array.map (fun hd -> Hashtbl.mem hd s) !map_lst in
    let find_var = Array.fold_left (fun x y -> x || y) false found_var_arr in
    if find_var = false then
      let local = L.build_alloca (ltype_of_typ t) s builder in
      Hashtbl.add (Array.get !map_lst 0) s local
  in
  let set_var s v =
    let hd = Array.get !map_lst 0 in
    Hashtbl.add hd s v
  in 
  let string_from_expr = function
    Sast.ANumLit (f, _) -> string_of_float f
  | Sast.ABoolLit (b, _) -> string_of_bool b
  | Sast.AStringLit (s, _) -> s
  | Sast.AVal (s, _) -> s
  | _ -> raise (failwith "Expression cannot be converted to string") in
  let get_type_from_list d = match d with
    | Sast.List(d) -> d
    | _ -> raise (failwith "not a list")

  in let rec expr builder = function
      Sast.ANumLit (i, _) -> L.const_float f_t i
    | Sast.ABoolLit (b, _) -> L.const_int i1_t (if b then 1 else 0)
    | Sast.AStringLit (s, _) -> L.build_global_stringptr s "string" builder
    | Sast.AVal (s, t) ->  (match t with
            Sast.List(_) -> lookup s 0
            | Sast.Object(_) -> lookup s 0
            | _ -> L.build_load (lookup s 0) s builder)
    | Sast.ABinop (e1, op, e2, t) ->
        let e1' = expr builder e1 in
        let e2' = expr builder e2
        and float_ops = (match op with
            Ast.Add     -> L.build_fadd
          | Ast.Minus   -> L.build_fsub
          | Ast.Times   -> L.build_fmul
          | Ast.Divide  -> L.build_fdiv
          | Ast.And     -> L.build_and
          | Ast.Or      -> L.build_or
          | Ast.Eq      -> L.build_fcmp L.Fcmp.Oeq
          | Ast.Neq     -> L.build_fcmp L.Fcmp.One
          | Ast.Less    -> L.build_fcmp L.Fcmp.Olt
          | Ast.Leq     -> L.build_fcmp L.Fcmp.Ole
          | Ast.Greater -> L.build_fcmp L.Fcmp.Ogt
          | Ast.Geq     -> L.build_fcmp L.Fcmp.Oge
          | _ -> L.build_fcmp L.Fcmp.Oeq
        )
        and int_ops = (match op with
            Ast.Add     -> L.build_add
          | Ast.Minus   -> L.build_sub
          | Ast.Times   -> L.build_mul
          | Ast.Divide  -> L.build_sdiv
          | Ast.And     -> L.build_and
          | Ast.Or      -> L.build_or
          | Ast.Eq      -> L.build_icmp L.Icmp.Eq
          | Ast.Neq     -> L.build_icmp L.Icmp.Ne
          | Ast.Less    -> L.build_icmp L.Icmp.Slt
          | Ast.Leq     -> L.build_icmp L.Icmp.Sle
          | Ast.Greater -> L.build_icmp L.Icmp.Sgt
          | Ast.Geq     -> L.build_icmp L.Icmp.Sge
          | _ -> L.build_icmp L.Icmp.Eq
        )
        and str_ops = (match op with
            Ast.Concat -> expr builder (Sast.AStringLit((string_from_expr e1) ^ (string_from_expr e2), t))
          | _ -> (L.const_int i32_t 0)
        )
        in
        if ((L.type_of e1' = str_t) && (L.type_of e2' = str_t)) then str_ops
        else if ((L.type_of e1' = f_t) && (L.type_of e2' = f_t)) then float_ops e1' e2' "tmp" builder
        else int_ops e1' e2' "tmp" builder
    | Sast.AUnop(op, e, _) ->
        let e' = expr builder e in
        (match op with
            Ast.Neg     -> L.build_fneg
          | Ast.Not     -> L.build_not) e' "tmp" builder
    | Sast.AAssign (s, e, t) -> 
      let e' = expr builder e in
      let _ = (match e with
        | Sast.AListDecl(ael, _) ->
          let _ = Hashtbl.add list_len_hash s (List.length ael) in ()
        | _ -> ())
      in (match t with 
        | Sast.Object(_) -> 
               let _ = set_var s e' in
               let _ = Hashtbl.add obj_ptr_map s (lookup s 0) in e'
        | Sast.List(_) -> 
                let _ = set_var s e' in e'
        | _ ->  let _ = add_var s t builder in
                let _ = L.build_store e' (lookup s 0) builder in e')
    | Sast.ACall (s, el, t) ->
        let (fdef, _) = StringMap.find s function_decls in
        let actuals = List.rev (List.map (expr builder) (List.rev el)) in
        let result = (match t with
          Sast.Void -> ""
        | _ -> s ^ "_result") in
        L.build_call fdef (Array.of_list actuals) result builder
    | Sast.ANoexpr _ -> L.const_int i32_t 0
    | Sast.AObjectAssign (e1, s, e2, t) ->
        let (e2', _) = 
          (match t with
            | Sast.Object(_) ->
              (match e2 with 
                | Sast.AVal(s, t) -> ((Hashtbl.find obj_ptr_map s), t)
                | _ -> ((expr builder e2), t)
              )
            | _ -> ((expr builder e2), t)
          )
        in
        let obj_name, obj_type = match e1 with
          Sast.AVal(s, t) -> (s, t)
        | _ -> raise (failwith "Not an object variable") in
        let obj_id = match obj_type with
          Sast.Object(id) -> id
        | _ -> -1 in
        let _ =
          if is_new_var obj_name 0 then
            let _ = add_var obj_name obj_type builder in
            let _ = Hashtbl.add obj_ptr_map obj_name (lookup obj_name 0) in
            let field_list = Hashtbl.find obj_hash obj_id in
            List.iter (fun (field_name, field_type) ->
              let e = (match field_type with
                | Sast.Num -> Sast.ANumLit(0.0, Sast.Num)
                | Sast.Bool -> Sast.ABoolLit(false, Sast.Bool)
                | Sast.String -> Sast.AStringLit("", Sast.String)
                | Sast.List(t) -> Sast.AListDecl([], Sast.List(t))
                | Sast.Object(_) -> Sast.ANoexpr(Sast.Num)
                | _ -> raise (failwith "Invalid object field type"))
              in
              if e = Sast.ANoexpr(Sast.Num)
              then ()
              else
                let e' = expr builder e in
                let var_name = (string_of_int obj_id) ^ "." ^ field_name in
                let field_idx = Hashtbl.find struct_field_indexes var_name in
                let _val = L.build_struct_gep (lookup obj_name 0) field_idx var_name builder in
                ignore(L.build_store e' _val builder)
            ) field_list
          else ()
        in
        let var_name = (string_of_int obj_id) ^ "." ^ s in
        let field_idx = Hashtbl.find struct_field_indexes var_name in
        let _ = add_var obj_name obj_type builder in
        let _ = Hashtbl.add obj_ptr_map obj_name (lookup obj_name 0) in
        let _val = L.build_struct_gep (lookup obj_name 0) field_idx var_name builder in
        let _ = L.build_store e2' _val builder in
        e2'
    | Sast.AObjectField (e, s, _) ->
        let obj_ptr, obj_type = (match e with 
          | Sast.AVal(s_2, t_2) -> 
          if (Hashtbl.mem obj_ptr_map s_2) then ((Hashtbl.find obj_ptr_map s_2), t_2)
                                      else ((expr builder e), t_2)             
          | Sast.AObjectField(_, _, t_1) ->  
            let e' = (expr builder e) in (e', t_1)
          | Sast.AListGet(_, _, t) ->
            let e' = (expr builder e) in (e', t)
          | _ -> raise (failwith "not yet supported"))
        in 
        let obj_id = (match obj_type with
          | Sast.Object(id) -> (string_of_int id)
          | _ -> raise (failwith "expected object")
        )
        in
        let search_term = (obj_id ^ "." ^ s) in
        let field_idx = Hashtbl.find struct_field_indexes search_term in
        let _val = L.build_struct_gep obj_ptr field_idx search_term builder in
        L.build_load _val s builder
    | Sast.AListDecl(el, d) ->
        let t = get_type_from_list d in
        let initialize_arr p el =
            let map_build x o =
                let x' = expr builder x in
                let arr_ptr = L.build_gep p [| L.const_int i32_t o |]
                    "tmp" builder in
                let _ = L.build_store x' arr_ptr builder
                in o + 1
            in List.fold_left (fun o e -> map_build e o) 0 el in

        let typ = match t with
          | Sast.Object(_) -> L.pointer_type (ltype_of_typ t)
          | Sast.List(u_t) -> L.pointer_type (ltype_of_typ u_t) 
          | _ -> ltype_of_typ t
        in
        let local = L.build_array_alloca typ
            (L.const_int i32_t (List.length el)) "local" builder
        in let _ = initialize_arr local el in
        local 
    | Sast.AListGet(e1, e2, _) ->
            let e1' = expr builder e1 in
            let e2' = expr builder e2 in
            let e2'' = L.build_fptoui e2' i32_t "tmp" builder in
            let pointer = L.build_gep e1' [|e2''|] "tmp" builder in
            L.build_load pointer "tmp" builder
    |  Sast.AListSet(e1, e2, e3, _) ->
            let e1' = expr builder e1 in
            let e2' = expr builder e2 in
            let e3' = expr builder e3 in
            let e2'' = L.build_fptoui e2' i32_t "tmp" builder in
            let pointer = L.build_gep e1' [|e2''|] "tmp" builder in
            let _ = L.build_store e3' pointer builder in
            L.build_load pointer "tmp" builder
    | Sast.AListLength(e1, _) ->
        (match e1 with
          | Sast.AVal(s, _) -> L.const_float f_t (float_of_int (Hashtbl.find
          list_len_hash s))
          | _ -> L.const_int i32_t 0)
    | Sast.AListPush(e1, e2, _) ->
        let old_size = (match e1 with
          | Sast.AVal(s, _) -> L.const_int i32_t (Hashtbl.find list_len_hash s)
          | _ -> L.const_int i32_t 0)
        in
        let load_values old_arr new_arr final_val arr_len start_pos builder =
          let new_block label =
            let f = L.block_parent (L.insertion_block builder) in
            L.append_block context label f
          in
          let bbcurr = L.insertion_block builder in
          let bbcond = new_block "array.cond" in
          let bbbody = new_block "array.init" in
          let bbdone = new_block "array.done" in
          ignore (L.build_br bbcond builder);
          L.position_at_end bbcond builder;

          (* Counter into the length of the array *)
          let counter = L.build_phi [L.const_int i32_t start_pos, bbcurr] "counter" builder in
          L.add_incoming ((L.build_add counter (L.const_int i32_t 1) "tmp" builder), bbbody) counter;
          let cmp = L.build_icmp L.Icmp.Slt counter arr_len "tmp1" builder in
          ignore (L.build_cond_br cmp bbbody bbdone builder);
          L.position_at_end bbbody builder;

          (* Assign array position to init_val *)
          let new_arr_ptr = L.build_gep new_arr [| counter |] "tmp2" builder in
          let old_arr_ptr = L.build_gep old_arr [| counter |] "tmp3" builder in
          let old_val = L.build_load old_arr_ptr "tmp4" builder in
          ignore (L.build_store old_val new_arr_ptr builder);
          ignore (L.build_br bbcond builder);

          L.position_at_end bbdone builder;
          let new_arr_ptr = L.build_gep new_arr [| counter |] "tmp5" builder in
          ignore (L.build_store final_val new_arr_ptr builder);

        in
        let e1' = expr builder e1 and e2' = expr builder e2 in
        let new_size = L.build_add old_size (L.const_int i32_t 1) "new_size" builder in
        let new_arr = L.build_array_alloca (L.type_of e2') new_size "tmp8" builder in
        let _ = load_values e1' new_arr e2' old_size 0 builder in
        let _ = match e1 with
          | Sast.AVal(s, _) -> let _ = set_var s new_arr in 
                let new_length = (Hashtbl.find list_len_hash s) in
                let _ = Hashtbl.replace list_len_hash s (new_length + 1) in
                ()
          | _ -> ()
        in e2'
    | Sast.AListDequeue(e, _) ->
        let e' = expr builder e in
        let new_arr= L.build_gep e' [|L.const_int i32_t 1 |] "tmp" builder in
        let _ = match e with
            | Sast.AVal(s, _) -> let _ = set_var s new_arr in
                 let new_length = (Hashtbl.find list_len_hash s) in
                 let _ = Hashtbl.replace list_len_hash s (new_length -1) in ()
            | _ -> ()
        in L.build_load e' "tmp" builder
    | Sast.AListPop(e, _) -> 
            let list_name = 
                (match e with
                  | Sast.AVal(s, _) -> s
                  | _ -> raise(Failure("pop on undefined list")))
            in
            let e' = expr builder e in 
            let list_length = (Hashtbl.find list_len_hash list_name) in  
            let _ = Hashtbl.replace list_len_hash list_name (list_length - 1) in
            let lst_ptr = L.build_gep e' [| L.const_int i32_t (list_length - 1) |] "tmp" builder
            in L.build_load lst_ptr "tmp" builder
    | Sast.AListRemove(e1, e2, d) ->
        let list_name = 
            (match e1 with
              | Sast.AVal(s, _) -> s
              | _ -> raise(Failure("Remove on undefined list")))
        in
        let e1' = expr builder e1 in
        let e2' = 
            (match e2 with
              | Sast.ANumLit(n, _) -> int_of_float n
              | _ -> raise (failwith "expressions not yet supported")
            )
        in
        let rec copy_w_rm new_arr old_arr old_ctr new_ctr rm_index length = 

        if (old_ctr >= length) then new_arr
        else
            (let new_arr_offset = L.build_gep new_arr [| L.const_int i32_t new_ctr |] "tmp3" builder in
            if (old_ctr == rm_index) then
                copy_w_rm new_arr old_arr  (old_ctr + 1) (new_ctr) rm_index length
            else 
            (
                let old_arr_offset = L.build_gep old_arr [| L.const_int i32_t old_ctr |] "tmp1" builder in
                let old_val = L.build_load old_arr_offset "tmp2" builder in
                let _ = L.build_store old_val new_arr_offset builder in
                copy_w_rm new_arr old_arr (old_ctr + 1) (new_ctr + 1) rm_index length
            ))
        in
        let new_list_length = ((Hashtbl.find list_len_hash list_name) - 1) in
        let local = L.build_array_alloca (ltype_of_typ d) 
            (L.const_int i32_t (new_list_length)) "tmp" builder
        in let new_arr_addr  = copy_w_rm local e1' 0 0 e2' (Hashtbl.find list_len_hash list_name) in 
        let _ = Hashtbl.replace list_len_hash list_name new_list_length in
        let _ = match e1 with
            | Sast.AVal(s, _) -> let _ = set_var s new_arr_addr in ()
            | _ -> ()
        in local
    | Sast.AListInsert(e1, e2, e3, d) ->
        let list_name = 
            (match e1 with
              | Sast.AVal(s, _) -> s
              | _ -> raise(Failure("Insert into undefined list")))
        in
        let e1' = expr builder e1 in
        let e2' = 
            (match e2 with
              | Sast.ANumLit(n, _) -> int_of_float n
              | _ -> raise (failwith "Expressions not yet supported"))
        in
        let e3' = expr builder e3 in
        let rec copy_w_ins new_arr old_arr ins_val old_ctr new_ctr ins_index length = 

        if (old_ctr >= length) then new_arr
        else
            (let new_arr_offset = L.build_gep new_arr [| L.const_int i32_t
            new_ctr |] "tmp3" builder in
            if (new_ctr == ins_index) then 
                let _ = L.build_store ins_val new_arr_offset builder in
                copy_w_ins new_arr old_arr ins_val (old_ctr) (new_ctr + 1) ins_index length
            else 
            (
                let old_arr_offset = L.build_gep old_arr [| L.const_int i32_t old_ctr |] "tmp1" builder in
                let old_val = L.build_load old_arr_offset "tmp2" builder in
                let _ = L.build_store old_val new_arr_offset builder in
                copy_w_ins new_arr old_arr ins_val (old_ctr + 1) (new_ctr + 1) ins_index length
            ))
        in
        let new_list_length = ((Hashtbl.find list_len_hash list_name) + 1) in
        let local = L.build_array_alloca (ltype_of_typ d) 
            (L.const_int i32_t (new_list_length)) "tmp" builder
        in let new_arr_addr  = copy_w_ins local e1' e3' 0 0 e2' (Hashtbl.find list_len_hash list_name) in 
        let _ = Hashtbl.replace list_len_hash list_name new_list_length in
        let _ = match e1 with
            | Sast.AVal(s,_) -> let _ = set_var s new_arr_addr in()
            | _ -> ()
        in e3'
    | _ -> L.const_float f_t 0.0
  in
  let add_terminal builder f =
    match L.block_terminator (L.insertion_block builder) with
        Some _ -> ()
      | None -> ignore (f builder);
  in

  let rec stmt builder = function
      Sast.ABlock sl -> List.fold_left stmt builder sl
    | Sast.AExpr e -> ignore (expr builder e); builder
    | Sast.AReturn e -> ignore (match fdecl.Sast.return with
        Sast.Void -> L.build_ret_void builder
      | _ -> L.build_ret (expr builder e) builder); builder
    | Sast.AIf (predicate, then_stmt, else_stmt) ->
      let bool_val = expr builder predicate in
      let merge_bb = L.append_block context "merge" the_function in

      map_lst := Array.append [|Hashtbl.create 100|] !map_lst;

      let then_bb = L.append_block context "then" the_function in
      add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
        (L.build_br merge_bb);

      map_lst := Array.sub !map_lst 1 ((Array.length !map_lst)-1);
      map_lst := Array.append [|Hashtbl.create 100|] !map_lst;

      let else_bb = L.append_block context "else" the_function in
      add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
        (L.build_br merge_bb);

      map_lst := Array.sub !map_lst 1 ((Array.length !map_lst)-1);

      ignore (L.build_cond_br bool_val then_bb else_bb builder);
      L.builder_at_end context merge_bb
    | Sast.AWhile (predicate, body) ->
        let pred_bb = L.append_block context "while" the_function in
        ignore (L.build_br pred_bb builder);
        let body_bb = L.append_block context "while_body" the_function in

        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = expr pred_builder predicate in

        let merge_bb = L.append_block context "merge" the_function in 
        map_lst := Array.append [|Hashtbl.create 100|] !map_lst;

        add_terminal (stmt (L.builder_at_end context body_bb) body)
          (L.build_br pred_bb);

        map_lst := Array.sub !map_lst 1 ((Array.length !map_lst)-1);

        ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb
    | Sast.AForIn (e1, e2, body) ->
        let vname = (match e1 with
          | Sast.AVal(id, _) -> id
          | _ -> raise (failwith "Not a variable"))
        in let length, sast_t = (match e2 with
          | Sast.AVal(id, t) -> Hashtbl.find list_len_hash id, get_type_from_list t
          | Sast.AListDecl(el, t) -> List.length el, get_type_from_list t
          | _ -> 0, Sast.Num)
        in
        let _ = add_var vname sast_t builder in
        let e2' = expr builder e2 in
        let rec map_iter_body cntr len =
          if cntr < len then
            let offset = L.const_int i32_t cntr in
            let val_ptr = L.build_gep e2' [| offset |] "val_ptr" builder in
            let new_val = L.build_load val_ptr "new_val" builder in
            let _ = L.build_store new_val (lookup vname 0) builder in

            map_lst := Array.append [|Hashtbl.create 100|] !map_lst;
            let _ = stmt builder body
            in map_lst := Array.sub !map_lst 1 ((Array.length !map_lst)-1);

            map_iter_body (cntr + 1) len
          else
            builder
        in map_iter_body 0 length;


    | Sast.AForRange (e1, e2, e3, body) ->
      let varname = match e1 with
        | Sast.AAssign(id, _, _) -> id
        | _ -> raise (failwith "Not an assign")
      in let e1' = expr builder e1 and e2' = expr builder e2
      and e3' = expr builder e3 in
      let e1'_ptr = (lookup varname 0) in
      let bool_val = L.build_fcmp L.Fcmp.Olt e1' e2' "lt" builder in
      let incr_ptr = L.build_alloca f_t "incr" builder in
      let _ = L.build_store e3' incr_ptr builder in    

      let new_block label =
        let f = L.block_parent (L.insertion_block builder) in
        L.append_block context label f
      in
      let _ = L.insertion_block builder in
      let incr_then = new_block "incr.incr" in
      let incr_else = new_block "incr.decr" in

      let bbcond = new_block "for.cond" in
      let bbbody = new_block "for.init" in
      let bbdone = new_block "for.done" in
      ignore (L.build_cond_br bool_val incr_then incr_else builder);

      (* increasing case*)
      L.position_at_end incr_then builder;
      let incr = L.build_load incr_ptr "incr" builder in
      let e1'' = L.build_fsub e1' incr "tmp" builder in
      let _ = L.build_store e1'' e1'_ptr builder in
      let _ = L.build_br bbcond builder in

      (* decreasing case *)
      L.position_at_end incr_else builder;
      let incr = L.build_load incr_ptr "incr" builder in
      let e1'' = L.build_fadd e1' incr "tmp" builder in
      let _ = L.build_store e1'' e1'_ptr builder in
      let _ = L.build_br bbcond builder in ();

      (* Iterate *)
      L.position_at_end bbcond builder;
      let load_itr = L.build_load e1'_ptr "iter" builder in
      let incr = L.build_load incr_ptr "incr" builder in
      let incr_itr = L.build_fadd load_itr incr "tmp" builder in
      let _ = L.build_store incr_itr e1'_ptr builder in
      let cmp = L.build_fcmp L.Fcmp.Olt incr_itr e2' "tmp1" builder in
      ignore (L.build_cond_br cmp bbbody bbdone builder);
      L.position_at_end bbbody builder;

      map_lst := Array.append [|Hashtbl.create 100|] !map_lst;

      add_terminal (stmt (L.builder_at_end context bbbody) body)
      (L.build_br bbcond);

      map_lst := Array.sub !map_lst 1 ((Array.length !map_lst)-1);

      L.position_at_end bbdone builder; builder

    | Sast.APrint e ->
      let get_field_name obj_field_access =
        let split = Str.split (Str.regexp "\\.") in
          List.hd (List.rev (split obj_field_access))
      in
      let e' = (expr builder e) in
      let e_type = L.string_of_lltype (L.type_of e') in

      let list_print iter ender list_len lptr =
        let get_load n =
          let e'_ptr = L.build_gep lptr [|L.const_int i32_t n|] "tmp" builder in
          L.build_load e'_ptr "tmp" builder
        in
        let rec iter_print ctr max acc = 
          if ctr = max then acc
          else let new_el = get_load ctr in iter_print (ctr + 1) max (acc @ [new_el])
        in
        let res = iter_print 0 list_len [] in 
        let rec print_list l =
          match l with
            | [] -> L.build_call printss_func
                [| expr builder (Sast.AStringLit("", Sast.String)) |] "puts" builder
            | [el] -> L.build_call printf_func [| ender; el|] "puts" builder
            | hd :: tail ->
                let _ = L.build_call printf_func [| iter; hd|] "puts" builder in
                print_list tail
        in
        let lbrace = expr builder (Sast.AStringLit("[", Sast.String)) in
        let rbrace = expr builder (Sast.AStringLit("]\n", Sast.String)) in
        let _ = L.build_call printss_func [| lbrace |] "puts" builder in
        let _ = print_list res in
        let _ = L.build_call printss_func [| rbrace |] "puts" builder in
        builder
      in
      if e_type = L.string_of_lltype f_t then
        (ignore(L.build_call printf_func [| float_format_str; e' |]
          "printf" builder); builder)
      else if e_type = L.string_of_lltype i1_t then
        (ignore(L.build_call printf_func [| int_format_str; e' |]
          "printf" builder); builder)
      else if e_type = L.string_of_lltype str_t then
        (ignore(L.build_call prints_func [|e'|]
          "puts" builder); builder)

      else if e_type = L.string_of_lltype (L.pointer_type str_t) then
        let length = match e with
          | Sast.AVal(s, _) -> Hashtbl.find list_len_hash s
          | Sast.AListDecl(el, _) -> List.length el
          | _ -> 0
        in
        list_print string_list_iter string_list_ender length e'

      else if e_type = L.string_of_lltype (L.pointer_type f_t) then
        let length = match e with
          | Sast.AVal(s, _) -> Hashtbl.find list_len_hash s
          | Sast.AListDecl(el, _) -> List.length el
          | _ -> 0
        in
        list_print float_list_iter float_list_ender length e'

      else if e_type = L.string_of_lltype (L.pointer_type i1_t) then
        let length = match e with
          | Sast.AVal(s, _) -> Hashtbl.find list_len_hash s
          | Sast.AListDecl(el, _) -> List.length el
          | _ -> 0
        in
        list_print int_list_iter int_list_ender length e'

      else
        let obj_name, obj_type = (match e with
          Sast.AVal(s, t) -> (s, t)
        | Sast.AObjectField(_, s, t) -> (s, t)
        | _ -> raise (failwith "Not a variable or an object variable") )in
        let obj_id = (match obj_type with
          Sast.Object(id) -> id
        | _ -> -1) in
        let split = Str.split (Str.regexp "\\.") in
        let obj_ptr = Hashtbl.find obj_ptr_map obj_name in
        let search_terms = Hashtbl.fold (fun k _ acc -> k :: acc) struct_field_indexes [] in
        let passed_terms = List.filter (fun term -> (List.hd (split term)) = (string_of_int obj_id)) search_terms in
        let field_idxs = List.map (fun k -> Hashtbl.find struct_field_indexes k) passed_terms in
        let val_arr = List.map2 (fun index field ->
          L.build_struct_gep obj_ptr index field builder) field_idxs passed_terms in
        let print_arr = List.map2 (fun struct_val field ->
          L.build_load struct_val (get_field_name field) builder) val_arr passed_terms
        in
        List.iter2 (fun x y ->
          let field_name = expr builder (Sast.AStringLit(obj_name ^ "." ^ (get_field_name y) ^ ": ", Sast.String)) in
          let e_type = L.string_of_lltype (L.type_of x) in
          if e_type = L.string_of_lltype f_t then
            (ignore(L.build_call printss_func [| field_name |] "puts" builder);
            ignore(L.build_call printf_func [| float_format_str; x |] "printf" builder))
          else if e_type = L.string_of_lltype i1_t then
            (ignore(L.build_call printss_func [| field_name |] "puts" builder);
            ignore(L.build_call printf_func [| int_format_str; x |] "printf" builder))
          else if e_type = L.string_of_lltype str_t then
            (ignore(L.build_call printss_func [| field_name |] "puts" builder);
            ignore(L.build_call prints_func [|x|] "puts" builder))
        ) print_arr passed_terms; builder
    | Sast.AObjectInit sl ->
        let _ = List.iter (fun (obj, obj_type) ->
          if not (Hashtbl.mem obj_ptr_map obj) then
            let obj_id = (match obj_type with
              Sast.Object(id) -> id
            | _ -> -1
            ) in
            let _ = add_var obj obj_type builder in
            let _ = Hashtbl.add obj_ptr_map obj (lookup obj 0) in
            let field_list = Hashtbl.find obj_hash obj_id in
            List.iter (fun (field_name, field_type) ->
              let e = (match field_type with
                  Sast.Num    -> Sast.ANumLit(0.0, Sast.Num)
                | Sast.Bool   -> Sast.ABoolLit(false, Sast.Bool)
                | Sast.String -> Sast.AStringLit("", Sast.String)
                | Sast.List(t) -> Sast.AListDecl([], Sast.List(t))
                | Sast.Object(_) -> Sast.ANoexpr(Sast.Num)
                | _ -> raise (failwith "Invalid object field type"))
              in
              if e = Sast.ANoexpr(Sast.Num)
              then ()
              else
                let e' = expr builder e in
                let var_name = (string_of_int obj_id) ^ "." ^ field_name in
                let field_idx = Hashtbl.find struct_field_indexes var_name in
                let _val = L.build_struct_gep (lookup obj 0) field_idx var_name builder in
                ignore(L.build_store e' _val builder)
            ) field_list
        ) sl in builder
    | _ -> builder
  in
  let builder = stmt builder (Sast.ABlock fdecl.Sast.abody) in
  add_terminal builder (match fdecl.Sast.return with
      Sast.String -> L.build_ret (L.build_global_stringptr "" "tmp" builder)
    | Sast.Num -> L.build_ret (L.const_float f_t 0.)
    | Sast.Bool -> L.build_ret (L.const_int i1_t 0)
    | Sast.Object(_) ->
      let arbitrary_obj = List.hd (Hashtbl.fold (fun _ v acc -> [v] @ acc) obj_ptr_map []) in
      L.build_ret arbitrary_obj
    | Sast.List(_) -> L.build_ret (expr builder (Sast.AListDecl([Sast.ANumLit(0., Sast.Num)], Sast.List(Sast.Num))))
    | _ -> L.build_ret_void)
in
List.iter build_function_body functions;
the_module
