/* Ocamlyacc parser for Pseudo */

%{
open Ast
%}

/* Punctuation */

%token PRINT
%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA DEF TILDE DOLLAR
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT
%token COMBINE TO CONCAT
%token EQ NEQ LT LEQ GT GEQ
%token AND OR IN BY
%token RETURN IF ELSE ELSIF FOR WHILE CONTINUE BREAK
%token LBRACKET RBRACKET COLON DOT
%token INSERT POP PUSH DEQUEUE REMOVE LENGTH GET SET
%token MEM FIND MAP DEL SIZE
%token INIT
%token <string> ID
%token EOF

/* Literals */
%token <float> NUM_LITERAL
%token <string> STRING_LITERAL
%token <bool> BOOL_LITERAL

%right ASSIGN
%left COMBINE
%left OR
%left AND
%left EQ NEQ IS
%left LT GT LEQ GEQ
%left PLUS MINUS CONCAT
%left TIMES DIVIDE
%right NOT
%nonassoc DOT

%start program
%type <Ast.program> program

%%
program:
	  func_decl_list EOF { List.rev $1 }

func_decl_list:
	/* nothing */ { [] }
	| func_decl_list func_decl {$2 :: $1}

obj_list:
	  ID { [$1] }
	| obj_list COMMA ID {$3 :: $1}

stmt:
	  expr SEMI {Expr $1}
	| WHILE expr TILDE loop_stmt_list DOLLAR {While($2, Block(List.rev $4))}
	| FOR ID IN expr TILDE loop_stmt_list DOLLAR {ForIn(Id($2), $4, Block(List.rev $6))}
	| FOR expr TO expr TILDE stmt_list DOLLAR {ForRange($2, $4, Num_lit(1.0), Block(List.rev $6))}
	| FOR expr TO expr BY expr TILDE stmt_list DOLLAR {ForRange($2, $4, $6, Block(List.rev $8))}
	| IF expr TILDE stmt_list DOLLAR elsif_list {If($2, Block(List.rev $4), Block([]))}
	| IF expr TILDE stmt_list DOLLAR elsif_list ELSE TILDE stmt_list DOLLAR {If($2, Block(List.rev $4), Block(List.rev $9))}
	| PRINT expr SEMI {Print($2)}
	| INIT obj_list SEMI {ObjectInit($2)}
	| RETURN SEMI {Return Noexpr}
	| RETURN expr SEMI {Return $2}	

stmt_list:
	  /* nothing */ { [] }
	| stmt_list stmt {$2 :: $1}

loop_stmt_list:
	  /* nothing */ { [] }
	| loop_stmt_list loop_stmt {$2 :: $1}

loop_stmt:
	  stmt { $1 }
	| BREAK SEMI { Break }
	| CONTINUE SEMI { Continue }

elsif_list:
	  /* nothing */ { [] }
	| elsif_list elsif_stmt {$2 :: $1}

elsif_stmt:
	| ELSIF expr TILDE stmt_list DOLLAR {If($2, Block(List.rev $4), Block([]))}

expr:
	  literal {$1}
	| arith_op {$1}
	| bool_op {$1}
	| string_op {$1}
	| LBRACKET pseudo_list RBRACKET {ListDecl(List.rev $2)}
	| list_op {$1}
	| dict_op {$1}
	| obj_op {$1}
	| ID {Id($1)}
	| ID ASSIGN expr {Assign($1, $3)}
	| LPAREN expr RPAREN {$2}
	| LBRACE dict RBRACE {DictDecl(List.rev $2)}
	| ID LPAREN actuals_opt RPAREN {Call($1, $3)}

arith_op:
	  MINUS expr		{Unop(Neg, $2)}
	| expr PLUS expr	{Binop($1, Add, $3)}
	| expr MINUS expr	{Binop($1, Minus, $3)}
	| expr TIMES expr	{Binop($1, Times, $3)}
	| expr DIVIDE expr	{Binop($1, Divide, $3)}

bool_op:
	  NOT expr		{Unop(Not, $2)}
	| expr EQ expr 	{Binop($1, Eq, $3)}
	| expr NEQ expr {Binop($1, Neq, $3)}
	| expr LT expr 	{Binop($1, Less, $3)}
	| expr LEQ expr {Binop($1, Leq, $3)}
	| expr GT expr 	{Binop($1, Greater, $3)}
	| expr GEQ expr {Binop($1, Geq, $3)}
	| expr AND expr {Binop($1, And, $3)}
	| expr OR expr	{Binop($1, Or, $3)}

string_op:
	  expr CONCAT expr {Binop($1, Concat, $3)}

pseudo_list:
	  /* nothing */ { [] }
	| expr { [$1] }
	| pseudo_list COMMA expr { $3 :: $1 }

dict: 
	  /* nothing */ { [] }
	| expr COLON expr { [($1, $3)] }
	| dict COMMA expr COLON expr { ($3, $5) :: $1 }

list_op:
	| expr COMBINE expr { Binop($1, Combine, $3) }
	| expr DOT INSERT LPAREN expr COMMA expr RPAREN { ListInsert($1, $5, $7) }
	| expr DOT PUSH LPAREN expr RPAREN { ListPush($1, $5) }
	| expr DOT POP LPAREN RPAREN { ListPop($1) }
	| expr DOT DEQUEUE LPAREN RPAREN { ListDequeue($1) }
	| expr DOT REMOVE LPAREN expr RPAREN { ListRemove($1, $5) }
	| expr DOT LENGTH LPAREN RPAREN { ListLength($1) }
	| expr DOT GET LPAREN expr RPAREN { ListGet($1, $5) }
	| expr DOT SET LPAREN expr COMMA expr RPAREN { ListSet($1, $5, $7) }
/*	| expr LBRACKET expr RBRACKET { ListGet($1, $3) }*/
/*	| expr LBRACKET expr RBRACKET ASSIGN expr { ListSet($1, $3, $6) }*/

dict_op:
	| expr DOT MEM LPAREN expr RPAREN { DictMem($1, $5) }
	| expr DOT FIND LPAREN expr RPAREN { DictFind($1, $5) }
	| expr DOT MAP LPAREN expr COMMA expr RPAREN { DictMap($1, $5, $7) }
	| expr DOT DEL LPAREN expr RPAREN { DictDelete($1, $5) }
	| expr DOT SIZE LPAREN RPAREN { DictSize($1) }
/*	| expr LBRACE expr RBRACE { DictFind($1, $3) }*/
/*	| expr LBRACE expr RBRACE ASSIGN expr { DictMap($1, $3, $6) }*/

obj_op:
    | expr DOT ID {ObjectField($1, $3)}
    | expr DOT ID ASSIGN expr {ObjectAssign($1, $3, $5)}

func_decl:
	  DEF ID LPAREN formal_opt RPAREN TILDE stmt_list DOLLAR
	  	{ {
	  		fname = $2;
	  		formals = $4;
	  		body = List.rev $7
	  	} }

actuals_opt:
	  /* nothing */ { [] }
	| actuals_list { List.rev $1 }

actuals_list:
      expr                    { [$1] }
  	| actuals_list COMMA expr { $3 :: $1 }

formal_opt:
	  /* nothing */ { [] }
	| formal_list { List.rev $1 }

formal_list:
	  ID { [$1] }
	| formal_list COMMA ID {$3 :: $1}

literal:
	| NUM_LITERAL 		{Num_lit($1)}
	| STRING_LITERAL	{String_lit($1)}
	| BOOL_LITERAL		{Bool_lit($1)}
